from typing import Any, Dict, List
from fluent.syntax import ast as FTL
from . import resolver


class Compiler:
    def __call__(self, item: Any) -> Any:
        if isinstance(item, FTL.BaseNode):
            return self.compile(item)
        if isinstance(item, (tuple, list)):
            return [self(elem) for elem in item]
        return item

    def compile(self, node: Any) -> Any:
        nodename: str = type(node).__name__
        if not hasattr(resolver, nodename):
            return node
        kwargs: Dict[str, Any] = vars(node).copy()
        for propname, propvalue in kwargs.items():
            kwargs[propname] = self(propvalue)
        handler = getattr(self, 'compile_' + nodename, self.compile_generic)
        return handler(nodename, **kwargs)

    def compile_generic(self, nodename: str, **kwargs: Any) -> Any:
        return getattr(resolver, nodename)(**kwargs)

    def compile_Placeable(self, _: Any, expression: Any, **kwargs: Any) -> Any:
        if isinstance(expression, resolver.Literal):
            return expression
        return resolver.Placeable(expression=expression, **kwargs)

    def compile_Pattern(self, _: Any, elements: List[Any], **kwargs: Any) -> Any:
        if len(elements) == 1 and isinstance(elements[0], resolver.Placeable):
            # Don't isolate isolated placeables
            return resolver.NeverIsolatingPlaceable(elements[0].expression)
        if any(
            not isinstance(child, resolver.Literal)
            for child in elements
        ):
            return resolver.Pattern(elements=elements, **kwargs)
        if len(elements) == 1:
            return elements[0]
        return resolver.TextElement(
            ''.join(child(None) for child in elements)
        )



import attr
import contextlib
from typing import Any, Dict, Generator, List, Set, TYPE_CHECKING, Union, cast

from fluent.syntax import ast as FTL
from .errors import FluentCyclicReferenceError, FluentFormatError, FluentReferenceError
from .types import FluentType, FluentNone, FluentInt, FluentFloat
from .utils import reference_to_id, unknown_reference_error_obj

if TYPE_CHECKING:
    from .bundle import FluentBundle


"""
The classes in this module are used to transform the source
AST to a partially evaluated resolver tree. They're subclasses
to the syntax AST node, and `BaseResolver`. Syntax nodes that
don't require special handling, but have children that need to be
transformed, need to just inherit from their syntax base class and
`BaseResolver`. When adding to the module namespace here, watch
out for naming conflicts with `fluent.syntax.ast`.

`ResolverEnvironment` is the `env` passed to the `__call__` method
in the resolver tree. The `CurrentEnvironment` keeps track of the
modifyable state in the resolver environment.
"""


# Prevent expansion of too long placeables, for memory DOS protection
MAX_PART_LENGTH = 2500


@attr.s
class CurrentEnvironment:
    # The parts of ResolverEnvironment that we want to mutate (and restore)
    # temporarily for some parts of a call chain.

    # The values of attributes here must not be mutated, they must only be
    # swapped out for different objects using `modified` (see below).

    # For Messages, VariableReference nodes are interpreted as external args,
    # but for Terms they are the values explicitly passed using CallExpression
    # syntax. So we have to be able to change 'args' for this purpose.
    args: Dict[str, Any] = attr.ib(factory=dict)
    # This controls whether we need to report an error if a VariableReference
    # refers to an arg that is not present in the args dict.
    error_for_missing_arg: bool = attr.ib(default=True)


@attr.s
class ResolverEnvironment:
    context: 'FluentBundle' = attr.ib()
    errors: List[Exception] = attr.ib()
    part_count: int = attr.ib(default=0, init=False)
    active_patterns: Set[FTL.Pattern] = attr.ib(factory=set, init=False)
    current: CurrentEnvironment = attr.ib(factory=CurrentEnvironment)

    @contextlib.contextmanager
    def modified(self, **replacements: Any) -> Generator['ResolverEnvironment', None, None]:
        """
        Context manager that modifies the 'current' attribute of the
        environment, restoring the old data at the end.
        """
        # CurrentEnvironment only has args that we never mutate, so the shallow
        # copy returned by attr.evolve is fine (at least for now).
        old_current = self.current
        self.current = attr.evolve(old_current, **replacements)
        yield self
        self.current = old_current

    def modified_for_term_reference(self, args: Union[Dict[str, Any], None] = None) -> Any:
        return self.modified(args=args if args is not None else {},
                             error_for_missing_arg=False)


class BaseResolver:
    """
    Abstract base class of all partially evaluated resolvers.

    Subclasses should implement __call__, with a
    ResolverEnvironment as parameter. An exception are wrapper
    classes that don't show up in the evaluation, but need to
    be part of the compiled tree structure.
    """

    def __call__(self, env: ResolverEnvironment) -> Any:
        raise NotImplementedError


class Literal(BaseResolver):
    value: str


class Message(FTL.Entry, BaseResolver):
    id: 'Identifier'
    value: Union['Pattern', None]
    attributes: Dict[str, 'Pattern']

    def __init__(self,
                 id: 'Identifier',
                 value: Union['Pattern', None] = None,
                 attributes: Union[List['Attribute'], None] = None,
                 comment: Any = None,
                 **kwargs: Any):
        super().__init__(**kwargs)
        self.id = id
        self.value = value
        self.attributes = {attr.id.name: attr.value for attr in attributes} if attributes else {}


class Term(FTL.Entry, BaseResolver):
    id: 'Identifier'
    value: 'Pattern'
    attributes: Dict[str, 'Pattern']

    def __init__(self,
                 id: 'Identifier',
                 value: 'Pattern',
                 attributes: Union[List['Attribute'], None] = None,
                 comment: Any = None,
                 **kwargs: Any):
        super().__init__(**kwargs)
        self.id = id
        self.value = value
        self.attributes = {attr.id.name: attr.value for attr in attributes} if attributes else {}


class Pattern(FTL.Pattern, BaseResolver):
    # Prevent messages with too many sub parts, for CPI DOS protection
    MAX_PARTS = 1000

    elements: List[Union['TextElement', 'Placeable']]  # type: ignore

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    def __call__(self, env: ResolverEnvironment) -> Union[str, FluentNone]:
        if self in env.active_patterns:
            env.errors.append(FluentCyclicReferenceError("Cyclic reference"))
            return FluentNone()
        env.active_patterns.add(self)
        elements = self.elements
        remaining_parts = self.MAX_PARTS - env.part_count
        if len(self.elements) > remaining_parts:
            env.active_patterns.remove(self)
            raise ValueError("Too many parts in message (> {}), "
                             "aborting.".format(self.MAX_PARTS))
        retval = ''.join(
            resolve(element(env), env) for element in elements
        )
        env.part_count += len(elements)
        env.active_patterns.remove(self)
        return retval


def resolve(fluentish: Any, env: ResolverEnvironment) -> Any:
    if isinstance(fluentish, FluentType):
        return fluentish.format(env.context._babel_locale)
    if isinstance(fluentish, str):
        if len(fluentish) > MAX_PART_LENGTH:
            raise ValueError(
                "Too many characters in placeable "
                "({}, max allowed is {})".format(len(fluentish), Pattern.MAX_PARTS)
            )
    return fluentish


class TextElement(FTL.TextElement, Literal):
    value: str

    def __call__(self, env: ResolverEnvironment) -> str:
        return self.value


class Placeable(FTL.Placeable, BaseResolver):
    expression: Union['InlineExpression', 'Placeable', 'SelectExpression']

    def __call__(self, env: ResolverEnvironment) -> Any:
        inner = resolve(self.expression(env), env)
        if not env.context.use_isolating:
            return inner
        return "\u2068" + inner + "\u2069"


class NeverIsolatingPlaceable(FTL.Placeable, BaseResolver):
    expression: Union['InlineExpression', Placeable, 'SelectExpression']

    def __call__(self, env: ResolverEnvironment) -> Any:
        inner = resolve(self.expression(env), env)
        return inner


class StringLiteral(FTL.StringLiteral, Literal):
    value: str

    def __call__(self, env: ResolverEnvironment) -> str:
        return self.parse()['value']


class NumberLiteral(FTL.NumberLiteral, BaseResolver):
    value: Union[FluentFloat, FluentInt]  # type: ignore

    def __init__(self, value: str, **kwargs: Any):
        super().__init__(value, **kwargs)
        if '.' in cast(str, self.value):
            self.value = FluentFloat(self.value)
        else:
            self.value = FluentInt(self.value)

    def __call__(self, env: ResolverEnvironment) -> Union[FluentFloat, FluentInt]:
        return self.value


def resolveEntryReference(
        ref: Union['MessageReference', 'TermReference'],
        env: ResolverEnvironment
) -> Union[str, FluentNone]:
    try:
        entry = env.context._lookup(ref.id.name, term=isinstance(ref, FTL.TermReference))
        pattern: Pattern
        if ref.attribute:
            pattern = entry.attributes[ref.attribute.name]
        else:
            pattern = entry.value  # type: ignore
        return pattern(env)
    except LookupError:
        ref_id = reference_to_id(ref)
        env.errors.append(unknown_reference_error_obj(ref_id))
        return FluentNone(f'{{{ref_id}}}')
    except TypeError:
        ref_id = reference_to_id(ref)
        env.errors.append(FluentReferenceError(f"No pattern: {ref_id}"))
        return FluentNone(ref_id)


class MessageReference(FTL.MessageReference, BaseResolver):
    id: 'Identifier'
    attribute: Union['Identifier', None]

    def __call__(self, env: ResolverEnvironment) -> Union[str, FluentNone]:
        return resolveEntryReference(self, env)


class TermReference(FTL.TermReference, BaseResolver):
    id: 'Identifier'
    attribute: Union['Identifier', None]
    arguments: Union['CallArguments', None]

    def __call__(self, env: ResolverEnvironment) -> Union[str, FluentNone]:
        if self.arguments:
            if self.arguments.positional:
                env.errors.append(FluentFormatError("Ignored positional arguments passed to term '{}'"
                                                    .format(reference_to_id(self))))
            kwargs = {kwarg.name.name: kwarg.value(env) for kwarg in self.arguments.named}
        else:
            kwargs = None
        with env.modified_for_term_reference(args=kwargs):
            return resolveEntryReference(self, env)


class VariableReference(FTL.VariableReference, BaseResolver):
    id: 'Identifier'

    def __call__(self, env: ResolverEnvironment) -> Any:
        name = self.id.name
        try:
            arg_val = env.current.args[name]
        except LookupError:
            if env.current.error_for_missing_arg:
                env.errors.append(
                    FluentReferenceError(f"Unknown external: {name}"))
            return FluentNone(name)

        if isinstance(arg_val, (FluentType, str)):
            return arg_val
        env.errors.append(TypeError("Unsupported external type: {}, {}"
                                    .format(name, type(arg_val))))
        return FluentNone(name)


class Attribute(FTL.Attribute, BaseResolver):
    id: 'Identifier'
    value: Pattern


class SelectExpression(FTL.SelectExpression, BaseResolver):
    selector: 'InlineExpression'
    variants: List['Variant']  # type: ignore

    def __call__(self, env: ResolverEnvironment) -> Union[str, FluentNone]:
        key = self.selector(env)
        default: Union['Variant', None] = None
        found: Union['Variant', None] = None
        for variant in self.variants:
            if variant.default:
                default = variant

            if match(key, variant.key(env), env):
                found = variant
                break

        if found is None:
            if default is None:
                env.errors.append(FluentFormatError("No default"))
                return FluentNone()
            found = default
        return found.value(env)


def is_number(val: Any) -> bool:
    return isinstance(val, (int, float))


def match(val1: Any, val2: Any, env: ResolverEnvironment) -> bool:
    if val1 is None or isinstance(val1, FluentNone):
        return False
    if val2 is None or isinstance(val2, FluentNone):
        return False
    if is_number(val1):
        if not is_number(val2):
            # Could be plural rule match
            return cast(bool, env.context._plural_form(val1) == val2)
    elif is_number(val2):
        return match(val2, val1, env)

    return cast(bool, val1 == val2)


class Variant(FTL.Variant, BaseResolver):
    key: Union['Identifier', NumberLiteral]
    value: Pattern
    default: bool


class Identifier(FTL.Identifier, BaseResolver):
    name: str

    def __call__(self, env: ResolverEnvironment) -> str:
        return self.name


class CallArguments(FTL.CallArguments, BaseResolver):
    positional: List[Union['InlineExpression', Placeable]]  # type: ignore
    named: List['NamedArgument']  # type: ignore


class FunctionReference(FTL.FunctionReference, BaseResolver):
    id: Identifier
    arguments: CallArguments

    def __call__(self, env: ResolverEnvironment) -> Any:
        args = [arg(env) for arg in self.arguments.positional]
        kwargs = {kwarg.name.name: kwarg.value(env) for kwarg in self.arguments.named}
        function_name = self.id.name
        try:
            function = env.context._functions[function_name]
        except LookupError:
            env.errors.append(FluentReferenceError("Unknown function: {}"
                                                   .format(function_name)))
            return FluentNone(function_name + "()")

        try:
            return function(*args, **kwargs)
        except Exception as e:
            env.errors.append(e)
            return FluentNone(function_name + "()")


class NamedArgument(FTL.NamedArgument, BaseResolver):
    name: Identifier
    value: Union[NumberLiteral, StringLiteral]


InlineExpression = Union[NumberLiteral, StringLiteral, MessageReference,
                         TermReference, VariableReference, FunctionReference]



from typing import cast


class FluentFormatError(ValueError):
    def __eq__(self, other: object) -> bool:
        return ((other.__class__ == self.__class__) and cast(ValueError, other).args == self.args)


class FluentReferenceError(FluentFormatError):
    pass


class FluentCyclicReferenceError(FluentFormatError):
    pass



from typing import Any, Callable, Dict
from .types import FluentType, fluent_date, fluent_number

NUMBER = fluent_number
DATETIME = fluent_date


BUILTINS: Dict[str, Callable[[Any], FluentType]] = {
    'NUMBER': NUMBER,
    'DATETIME': DATETIME,
}



import babel
import babel.numbers
import babel.plural
from typing import Any, Callable, Dict, List, TYPE_CHECKING, Tuple, Union, cast
from typing_extensions import Literal

from fluent.syntax import ast as FTL

from .builtins import BUILTINS
from .prepare import Compiler
from .resolver import CurrentEnvironment, Message, Pattern, ResolverEnvironment
from .utils import native_to_fluent

if TYPE_CHECKING:
    from .types import FluentNone, FluentType

PluralCategory = Literal['zero', 'one', 'two', 'few', 'many', 'other']


class FluentBundle:
    """
    Bundles are single-language stores of translations.  They are
    aggregate parsed Fluent resources in the Fluent syntax and can
    format translation units (entities) to strings.

    Always use `FluentBundle.get_message` to retrieve translation units from
    a bundle. Generate the localized string by using `format_pattern` on
    `message.value` or `message.attributes['attr']`.
    Translations can contain references to other entities or
    external arguments, conditional logic in form of select expressions, traits
    which describe their grammatical features, and can use Fluent builtins.
    See the documentation of the Fluent syntax for more information.
    """

    def __init__(self,
                 locales: List[str],
                 functions: Union[Dict[str, Callable[[Any], 'FluentType']], None] = None,
                 use_isolating: bool = True):
        self.locales = locales
        self._functions = {**BUILTINS, **(functions or {})}
        self.use_isolating = use_isolating
        self._messages: Dict[str, Union[FTL.Message, FTL.Term]] = {}
        self._terms: Dict[str, Union[FTL.Message, FTL.Term]] = {}
        self._compiled: Dict[str, Message] = {}
        # The compiler is not typed, and this cast is only valid for the public API
        self._compiler = cast(Callable[[Union[FTL.Message, FTL.Term]], Message], Compiler())
        self._babel_locale = self._get_babel_locale()
        self._plural_form = cast(Callable[[Any], Callable[[Union[int, float]], PluralCategory]],
                                 babel.plural.to_python)(self._babel_locale.plural_form)

    def add_resource(self, resource: FTL.Resource, allow_overrides: bool = False) -> None:
        # TODO - warn/error about duplicates
        for item in resource.body:
            if not isinstance(item, (FTL.Message, FTL.Term)):
                continue
            map_ = self._messages if isinstance(item, FTL.Message) else self._terms
            full_id = item.id.name
            if full_id not in map_ or allow_overrides:
                map_[full_id] = item

    def has_message(self, message_id: str) -> bool:
        return message_id in self._messages

    def get_message(self, message_id: str) -> Message:
        return self._lookup(message_id)

    def _lookup(self, entry_id: str, term: bool = False) -> Message:
        if term:
            compiled_id = '-' + entry_id
        else:
            compiled_id = entry_id
        try:
            return self._compiled[compiled_id]
        except LookupError:
            pass
        entry = self._terms[entry_id] if term else self._messages[entry_id]
        self._compiled[compiled_id] = self._compiler(entry)
        return self._compiled[compiled_id]

    def format_pattern(self,
                       pattern: Pattern,
                       args: Union[Dict[str, Any], None] = None
                       ) -> Tuple[Union[str, 'FluentNone'], List[Exception]]:
        if args is not None:
            fluent_args = {
                argname: native_to_fluent(argvalue)
                for argname, argvalue in args.items()
            }
        else:
            fluent_args = {}

        errors: List[Exception] = []
        env = ResolverEnvironment(context=self,
                                  current=CurrentEnvironment(args=fluent_args),
                                  errors=errors)
        try:
            result = pattern(env)
        except ValueError as e:
            errors.append(e)
            result = '{???}'
        return (result, errors)

    def _get_babel_locale(self) -> babel.Locale:
        for lc in self.locales:
            try:
                return babel.Locale.parse(lc.replace('-', '_'))
            except babel.UnknownLocaleError:
                continue
        # TODO - log error
        return babel.Locale.default()



from datetime import date, datetime
from decimal import Decimal
from typing import Any, Union

from fluent.syntax.ast import MessageReference, TermReference

from .types import FluentInt, FluentFloat, FluentDecimal, FluentDate, FluentDateTime
from .errors import FluentReferenceError

TERM_SIGIL = '-'
ATTRIBUTE_SEPARATOR = '.'


def native_to_fluent(val: Any) -> Any:
    """
    Convert a python type to a Fluent Type.
    """
    if isinstance(val, int):
        return FluentInt(val)
    if isinstance(val, float):
        return FluentFloat(val)
    if isinstance(val, Decimal):
        return FluentDecimal(val)

    if isinstance(val, datetime):
        return FluentDateTime.from_date_time(val)
    if isinstance(val, date):
        return FluentDate.from_date(val)
    return val


def reference_to_id(ref: Union[MessageReference, TermReference]) -> str:
    """
    Returns a string reference for a MessageReference or TermReference
    AST node.

    e.g.
       message
       message.attr
       -term
       -term.attr
    """
    start: str
    if isinstance(ref, TermReference):
        start = TERM_SIGIL + ref.id.name
    else:
        start = ref.id.name

    if ref.attribute:
        return ''.join([start, ATTRIBUTE_SEPARATOR, ref.attribute.name])
    return start


def unknown_reference_error_obj(ref_id: str) -> FluentReferenceError:
    if ATTRIBUTE_SEPARATOR in ref_id:
        return FluentReferenceError(f"Unknown attribute: {ref_id}")
    if ref_id.startswith(TERM_SIGIL):
        return FluentReferenceError(f"Unknown term: {ref_id}")
    return FluentReferenceError(f"Unknown message: {ref_id}")



import warnings
from datetime import date, datetime
from decimal import Decimal

import attr
import pytz
from babel import Locale
from babel.dates import format_date, format_time, get_datetime_format, get_timezone
from babel.numbers import NumberPattern, parse_pattern
from typing import Any, Dict, Type, TypeVar, Union, cast
from typing_extensions import Literal

FORMAT_STYLE_DECIMAL = "decimal"
FORMAT_STYLE_CURRENCY = "currency"
FORMAT_STYLE_PERCENT = "percent"
FORMAT_STYLE_OPTIONS = {
    FORMAT_STYLE_DECIMAL,
    FORMAT_STYLE_CURRENCY,
    FORMAT_STYLE_PERCENT,
}

CURRENCY_DISPLAY_SYMBOL = "symbol"
CURRENCY_DISPLAY_CODE = "code"
CURRENCY_DISPLAY_NAME = "name"
CURRENCY_DISPLAY_OPTIONS = {
    CURRENCY_DISPLAY_SYMBOL,
    CURRENCY_DISPLAY_CODE,
    CURRENCY_DISPLAY_NAME,
}

DATE_STYLE_OPTIONS = {
    "full",
    "long",
    "medium",
    "short",
    None,
}

TIME_STYLE_OPTIONS = {
    "full",
    "long",
    "medium",
    "short",
    None,
}


class FluentType:
    def format(self, locale: Locale) -> str:
        raise NotImplementedError()


class FluentNone(FluentType):
    def __init__(self, name: Union[str, None] = None):
        self.name = name

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, FluentNone) and self.name == other.name

    def format(self, locale: Locale) -> str:
        return self.name or "???"


@attr.s
class NumberFormatOptions:
    # We follow the Intl.NumberFormat parameter names here,
    # rather than using underscores as per PEP8, so that
    # we can stick to Fluent spec more easily.

    # See https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/NumberFormat
    style: Literal['decimal', 'currency', 'percent'] = attr.ib(
        default=FORMAT_STYLE_DECIMAL,
        validator=attr.validators.in_(FORMAT_STYLE_OPTIONS))
    currency: Union[str, None] = attr.ib(default=None)
    currencyDisplay: Literal['symbol', 'code', 'name'] = attr.ib(
        default=CURRENCY_DISPLAY_SYMBOL,
        validator=attr.validators.in_(CURRENCY_DISPLAY_OPTIONS))
    useGrouping: bool = attr.ib(default=True)
    minimumIntegerDigits: Union[int, None] = attr.ib(default=None)
    minimumFractionDigits: Union[int, None] = attr.ib(default=None)
    maximumFractionDigits: Union[int, None] = attr.ib(default=None)
    minimumSignificantDigits: Union[int, None] = attr.ib(default=None)
    maximumSignificantDigits: Union[int, None] = attr.ib(default=None)


class FluentNumber(FluentType):

    default_number_format_options = NumberFormatOptions()

    def __new__(cls,
                value: Union[int, float, Decimal, 'FluentNumber'],
                **kwargs: Any) -> 'FluentNumber':
        self = super().__new__(cls, value)  # type: ignore
        return self._init(value, kwargs)

    def _init(self,
              value: Union[int, float, Decimal, 'FluentNumber'],
              kwargs: Dict[str, Any]) -> 'FluentNumber':
        self.options = merge_options(NumberFormatOptions,
                                     getattr(value, 'options', self.default_number_format_options),
                                     kwargs)

        if self.options.style == FORMAT_STYLE_CURRENCY and self.options.currency is None:
            raise ValueError("currency must be provided")

        return self

    def format(self, locale: Locale) -> str:
        selfnum = cast(float, self)
        if self.options.style == FORMAT_STYLE_DECIMAL:
            base_pattern = cast(NumberPattern, locale.decimal_formats.get(None))
            pattern = self._apply_options(base_pattern)
            return pattern.apply(selfnum, locale)
        elif self.options.style == FORMAT_STYLE_PERCENT:
            base_pattern = cast(NumberPattern, locale.percent_formats.get(None))
            pattern = self._apply_options(base_pattern)
            return pattern.apply(selfnum, locale)
        elif self.options.style == FORMAT_STYLE_CURRENCY:
            base_pattern = locale.currency_formats['standard']
            pattern = self._apply_options(base_pattern)
            return pattern.apply(selfnum, locale, currency=self.options.currency)
        # never happens
        return '???'

    def _apply_options(self, pattern: NumberPattern) -> NumberPattern:
        # We are essentially trying to copy the
        # https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/NumberFormat
        # API using Babel number formatting routines, which is slightly awkward
        # but not too bad as they are both based on Unicode standards.

        # The easiest route is to start from the existing NumberPattern, and
        # then change its attributes so that Babel's number formatting routines
        # do the right thing. The NumberPattern.pattern string then becomes
        # incorrect, but it is not used when formatting, it is only used
        # initially to set the other attributes.
        pattern = clone_pattern(pattern)
        if not self.options.useGrouping:
            pattern.grouping = _UNGROUPED_PATTERN.grouping
        if self.options.style == FORMAT_STYLE_CURRENCY:
            if self.options.currencyDisplay == CURRENCY_DISPLAY_CODE:
                # Not sure of the correct algorithm here, but this seems to
                # work:
                def replacer(s: str) -> str:
                    return s.replace("¤", "¤¤")
                pattern.suffix = (replacer(pattern.suffix[0]),
                                  replacer(pattern.suffix[1]))
                pattern.prefix = (replacer(pattern.prefix[0]),
                                  replacer(pattern.prefix[1]))
            elif self.options.currencyDisplay == CURRENCY_DISPLAY_NAME:
                # No support for this yet - see
                # https://github.com/python-babel/babel/issues/578 But it's
                # better to display something than crash or a generic fallback
                # string, so we just issue a warning and carry on for now.
                warnings.warn("Unsupported currencyDisplayValue {}, falling back to {}"
                              .format(CURRENCY_DISPLAY_NAME,
                                      CURRENCY_DISPLAY_SYMBOL))
        minSD = self.options.minimumSignificantDigits
        maxSD = self.options.maximumSignificantDigits
        if (minSD is not None or maxSD is not None):
            # This triggers babel routines into 'significant digits' mode:
            pattern.pattern = '@'
            # We then manually set int_prec, and leave the rest as they are.
            min_digits = minSD if minSD is not None else 1
            max_digits = maxSD if maxSD is not None else min_digits
            pattern.int_prec = (min_digits, max_digits)
        else:
            if self.options.minimumIntegerDigits is not None:
                pattern.int_prec = (self.options.minimumIntegerDigits, pattern.int_prec[1])
            if self.options.minimumFractionDigits is not None:
                pattern.frac_prec = (self.options.minimumFractionDigits, pattern.frac_prec[1])
            if self.options.maximumFractionDigits is not None:
                pattern.frac_prec = (pattern.frac_prec[0], self.options.maximumFractionDigits)

        return pattern


Options = TypeVar('Options', bound=Union[NumberFormatOptions, 'DateFormatOptions'])


def merge_options(options_class: Type[Options], base: Union[Options, None], kwargs: Dict[str, Any]) -> Options:
    """
    Given an 'options_class', an optional 'base' object to copy from,
    and some keyword arguments, create a new options instance
    """
    if base is not None and not kwargs:
        # We can safely re-use base, because we don't
        # mutate options objects outside this function.
        return base

    retval = options_class()

    if base is not None:
        # We only copy values in `__dict__` to avoid class attributes.
        retval.__dict__.update(base.__dict__)

    # Use the options_class constructor because it might
    # have validators defined for the fields.
    kwarg_options = options_class(**kwargs)
    # Then merge, using only the ones explicitly given as keyword params.
    for k in kwargs.keys():
        setattr(retval, k, getattr(kwarg_options, k))

    return retval  # type: ignore


# We want types that inherit from both FluentNumber and a native type,
# so that:
#
# 1) developers can just pass native types if they don't want to specify
#    options, and fluent should handle these the same internally.
#
# 2) if they are using functions in messages, these can be passed FluentNumber
#    instances in place of a native type and will work just the same without
#    modification (in most cases).

class FluentInt(FluentNumber, int):
    pass


class FluentFloat(FluentNumber, float):
    pass


class FluentDecimal(FluentNumber, Decimal):
    pass


def fluent_number(
        number: Union[int, float, Decimal, FluentNumber, FluentNone],
        **kwargs: Any
) -> Union[FluentNumber, FluentNone]:
    if isinstance(number, FluentNumber) and not kwargs:
        return number
    if isinstance(number, int):
        return FluentInt(number, **kwargs)
    elif isinstance(number, float):
        return FluentFloat(number, **kwargs)
    elif isinstance(number, Decimal):
        return FluentDecimal(number, **kwargs)
    elif isinstance(number, FluentNone):
        return number
    else:
        raise TypeError("Can't use fluent_number with object {} for type {}"
                        .format(number, type(number)))


_UNGROUPED_PATTERN = parse_pattern("#0")


def clone_pattern(pattern: NumberPattern) -> NumberPattern:
    return NumberPattern(pattern.pattern,
                         pattern.prefix,
                         pattern.suffix,
                         pattern.grouping,
                         pattern.int_prec,
                         pattern.frac_prec,
                         pattern.exp_prec,
                         pattern.exp_plus)


@attr.s
class DateFormatOptions:
    # Parameters.
    # See https://projectfluent.org/fluent/guide/functions.html#datetime

    # Developer only
    timeZone: Union[str, None] = attr.ib(default=None)

    # Other
    hour12: Union[bool, None] = attr.ib(default=None)
    weekday: Literal["long", "short", "narrow", None] = attr.ib(default=None)
    era: Literal["long", "short", "narrow", None] = attr.ib(default=None)
    year: Literal["numeric", "2-digit", None] = attr.ib(default=None)
    month: Literal["numeric", "2-digit", "long", "short", "narrow", None] = attr.ib(default=None)
    day: Literal["numeric", "2-digit", None] = attr.ib(default=None)
    hour: Literal["numeric", "2-digit", None] = attr.ib(default=None)
    minute: Literal["numeric", "2-digit", None] = attr.ib(default=None)
    second: Literal["numeric", "2-digit", None] = attr.ib(default=None)
    timeZoneName: Literal["long", "short", "longOffset", "shortOffset",
                          "longGeneric", "shortGeneric", None] = attr.ib(default=None)

    # See https://github.com/tc39/proposal-ecma402-datetime-style
    dateStyle: Literal["full", "long", "medium", "short", None] = attr.ib(
        default=None,
        validator=attr.validators.in_(DATE_STYLE_OPTIONS))
    timeStyle: Literal["full", "long", "medium", "short", None] = attr.ib(
        default=None,
        validator=attr.validators.in_(TIME_STYLE_OPTIONS))


_SUPPORTED_DATETIME_OPTIONS = ['dateStyle', 'timeStyle', 'timeZone']


class FluentDateType(FluentType):
    # We need to match signature of `__init__` and `__new__` due to the way
    # some Python implementation (e.g. PyPy) implement some methods.
    # So we leave those alone, and implement another `_init_options`
    # which is called from other constructors.
    def _init_options(self, dt_obj: Union[date, datetime], kwargs: Dict[str, Any]) -> None:
        if 'timeStyle' in kwargs and not isinstance(self, datetime):
            raise TypeError("timeStyle option can only be specified for datetime instances, not date instance")

        self.options = merge_options(DateFormatOptions,
                                     getattr(dt_obj, 'options', None),
                                     kwargs)
        for k in kwargs:
            if k not in _SUPPORTED_DATETIME_OPTIONS:
                warnings.warn(f"FluentDateType option {k} is not yet supported")

    def format(self, locale: Locale) -> str:
        if isinstance(self, datetime):
            selftz = _ensure_datetime_tzinfo(self, tzinfo=self.options.timeZone)
        else:
            selftz = cast(datetime, self)

        ds = self.options.dateStyle
        ts = self.options.timeStyle
        if ds is None:
            if ts is None:
                return format_date(selftz, format='medium', locale=locale)
            else:
                return format_time(selftz, format=ts, locale=locale)
        elif ts is None:
            return format_date(selftz, format=ds, locale=locale)

        # Both date and time. Logic copied from babel.dates.format_datetime,
        # with modifications.
        # Which datetime format do we pick? We arbitrarily pick dateStyle.

        return (cast(str, get_datetime_format(ds, locale=locale))
                .replace("'", "")
                .replace('{0}', format_time(selftz, ts, tzinfo=None, locale=locale))
                .replace('{1}', format_date(selftz, ds, locale=locale)))


def _ensure_datetime_tzinfo(dt: datetime, tzinfo: Union[str, None] = None) -> datetime:
    """
    Ensure the datetime passed has an attached tzinfo.
    """
    # Adapted from babel's function.
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=pytz.UTC)
    if tzinfo is not None:
        dt = dt.astimezone(get_timezone(tzinfo))
    return dt


class FluentDate(FluentDateType, date):
    @classmethod
    def from_date(cls, dt_obj: date, **kwargs: Any) -> 'FluentDate':
        obj = cls(dt_obj.year, dt_obj.month, dt_obj.day)
        obj._init_options(dt_obj, kwargs)
        return obj


class FluentDateTime(FluentDateType, datetime):
    @classmethod
    def from_date_time(cls, dt_obj: datetime, **kwargs: Any) -> 'FluentDateTime':
        obj = cls(dt_obj.year, dt_obj.month, dt_obj.day,
                  dt_obj.hour, dt_obj.minute, dt_obj.second,
                  dt_obj.microsecond, tzinfo=dt_obj.tzinfo)
        obj._init_options(dt_obj, kwargs)
        return obj


def fluent_date(
        dt: Union[date, datetime, FluentDateType, FluentNone],
        **kwargs: Any
) -> Union[FluentDateType, FluentNone]:
    if isinstance(dt, FluentDateType) and not kwargs:
        return dt
    if isinstance(dt, datetime):
        return FluentDateTime.from_date_time(dt, **kwargs)
    elif isinstance(dt, date):
        return FluentDate.from_date(dt, **kwargs)
    elif isinstance(dt, FluentNone):
        return dt
    else:
        raise TypeError("Can't use fluent_date with object {} of type {}"
                        .format(dt, type(dt)))



import codecs
import os
from typing import Any, Callable, Dict, Generator, List, TYPE_CHECKING, Type, Union, cast

from fluent.syntax import FluentParser

from .bundle import FluentBundle

if TYPE_CHECKING:
    from fluent.syntax.ast import Resource
    from .types import FluentType


class FluentLocalization:
    """
    Generic API for Fluent applications.

    This handles language fallback, bundle creation and string localization.
    It uses the given resource loader to load and parse Fluent data.
    """

    def __init__(
        self,
        locales: List[str],
        resource_ids: List[str],
        resource_loader: 'AbstractResourceLoader',
        use_isolating: bool = False,
        bundle_class: Type[FluentBundle] = FluentBundle,
        functions: Union[Dict[str, Callable[[Any], 'FluentType']], None] = None,
    ):
        self.locales = locales
        self.resource_ids = resource_ids
        self.resource_loader = resource_loader
        self.use_isolating = use_isolating
        self.bundle_class = bundle_class
        self.functions = functions
        self._bundle_cache: List[FluentBundle] = []
        self._bundle_it = self._iterate_bundles()

    def format_value(self, msg_id: str, args: Union[Dict[str, Any], None] = None) -> str:
        for bundle in self._bundles():
            if not bundle.has_message(msg_id):
                continue
            msg = bundle.get_message(msg_id)
            if not msg.value:
                continue
            val, _errors = bundle.format_pattern(msg.value, args)
            return cast(str, val)  # Never FluentNone when format_pattern called externally
        return msg_id

    def _create_bundle(self, locales: List[str]) -> FluentBundle:
        return self.bundle_class(
            locales, functions=self.functions, use_isolating=self.use_isolating
        )

    def _bundles(self) -> Generator[FluentBundle, None, None]:
        bundle_pointer = 0
        while True:
            if bundle_pointer == len(self._bundle_cache):
                try:
                    self._bundle_cache.append(next(self._bundle_it))
                except StopIteration:
                    return
            yield self._bundle_cache[bundle_pointer]
            bundle_pointer += 1

    def _iterate_bundles(self) -> Generator[FluentBundle, None, None]:
        for first_loc in range(0, len(self.locales)):
            locs = self.locales[first_loc:]
            for resources in self.resource_loader.resources(locs[0], self.resource_ids):
                bundle = self._create_bundle(locs)
                for resource in resources:
                    bundle.add_resource(resource)
                yield bundle


class AbstractResourceLoader:
    """
    Interface to implement for resource loaders.
    """

    def resources(self, locale: str, resource_ids: List[str]) -> Generator[List['Resource'], None, None]:
        """
        Yield lists of FluentResource objects, corresponding to
        each of the resource_ids.
        If there are multiple locations, this may yield multiple lists.
        If a resource isn't found in any location, yield a partial list,
        but don't yield empty lists.
        """
        raise NotImplementedError


class FluentResourceLoader(AbstractResourceLoader):
    """
    Resource loader to read Fluent files from disk.

    Different locales are in different locations based on locale code.
    The locale code should be encoded as `{locale}` in the roots, or in
    the resource_ids.
    This loader does not support loading resources for one bundle from
    different roots.
    """

    def __init__(self, roots: Union[str, List[str]]):
        """
        Create a resource loader. The roots may be a string for a single
        location on disk, or a list of strings.
        """
        self.roots = [roots] if isinstance(roots, str) else roots

    def resources(self, locale: str, resource_ids: List[str]) -> Generator[List['Resource'], None, None]:
        for root in self.roots:
            resources: List[Any] = []
            for resource_id in resource_ids:
                path = self.localize_path(os.path.join(root, resource_id), locale)
                if not os.path.isfile(path):
                    continue
                content = codecs.open(path, 'r', 'utf-8').read()
                resources.append(FluentParser().parse(content))
            if resources:
                yield resources

    def localize_path(self, path: str, locale: str) -> str:
        return path.format(locale=locale)



from fluent.syntax import FluentParser
from fluent.syntax.ast import Resource

from .bundle import FluentBundle
from .fallback import FluentLocalization, AbstractResourceLoader, FluentResourceLoader


__all__ = [
    'FluentLocalization',
    'AbstractResourceLoader',
    'FluentResourceLoader',
    'FluentResource',
    'FluentBundle',
]


def FluentResource(source: str) -> Resource:
    parser = FluentParser()
    return parser.parse(source)
